# SPDX-FileCopyrightText: © 2024-2025 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


class Duration:
    """A selection of set durations, in seconds."""

    ONE_CENTISECOND = 0.01
    ONE_DECISECOND = 0.1
    ONE_SECOND = 1.0
    TEN_SECONDS = 10.0
